"""
Module for querying astronomical sources by name
"""

import logging

import pandas as pd
from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.mast import Catalogs
from requests.exceptions import HTTPError

from galsynthspec.datamodels.galaxy import Galaxy
from galsynthspec.skyportal import client, query_skyportal_by_name
from galsynthspec.utils.tns import get_tns_by_name

logger = logging.getLogger(__name__)


def load_source_info(name: str, use_cache: bool = True) -> pd.Series:
    """
    Load source information from SkyPortal by name.

    :param name: The name of the astronomical source.
    :param use_cache: If True, use cached data if available.
    :return: A dictionary containing the source information.
    """
    if client.has_skyportal_token():
        if client.ping():
            try:
                return query_skyportal_by_name(name)
            except HTTPError as e:
                logger.debug(f"Error querying SkyPortal for {name}: {e}")
                logger.debug("Falling back to TNS query.")

    return get_tns_by_name(name, use_cache=use_cache)


def query_by_name(name: str, use_cache: bool = True) -> Galaxy:
    """
    Query the SkyPortal API for a source by ZTF or AT name.
    Create a Galaxy object centered on the nearest Panstarrs match.

    :param name: AT/SN/TDE or ZTF/ATLAS/Gaia etc name of the source.
    :param use_cache: If True, use cached data if available.
    :return: Galaxy object based on nearest Panstarrs host.
    """

    data = load_source_info(name, use_cache=use_cache)

    src_ra, src_dec = data["ra"], data["dec"]

    redshift = data["redshift"] if "redshift" in data else None

    src_position = SkyCoord(src_ra, src_dec, unit="deg")

    catalog_data = Catalogs.query_region(  # pylint: disable=no-member
        src_position,
        radius=10.0 * u.arcsec,  # pylint: disable=no-member
        catalog="Panstarrs",
    )

    if len(catalog_data) > 1:
        logger.warning(
            f"Multiple Panstarrs matches found for {name}. Will use the nearest one."
        )

    match = catalog_data.group_by("distance")[0]

    gal_ra, gal_dec = match["raMean"], match["decMean"]

    return Galaxy(source_name=name, ra_deg=gal_ra, dec_deg=gal_dec, redshift=redshift)



"""
Module for GALEX data
"""

# pylint: disable=duplicate-code

import logging

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.mast import Catalogs

from galsynthspec.datamodels.photometry import Photometry

GALEX_BANDS = ["FUV", "NUV"]
GALEX_MAG_COLS = [f"{x.lower()}_mag" for x in GALEX_BANDS]
GALEX_MAGERR_COLS = [f"{x.lower()}_magerr" for x in GALEX_BANDS]

logger = logging.getLogger(__name__)


def download_galex_data(
    src_position: SkyCoord,
    radius_arcsec: float,
) -> list[Photometry]:
    """
    Module for downloading PS1 data for a given galaxy.

    Parameters
    :param src_position: SkyCoord The position of the source in the sky.
    :param radius_arcsec: float The radius of the search in arcseconds.

    Returns
    :return: list[Photometry] The photometry data for the source.
    """

    all_filters = []

    catalog_data = Catalogs.query_region(  # pylint: disable=no-member
        src_position,
        radius=radius_arcsec * u.arcsec,  # pylint: disable=no-member
        catalog="Galex",
    )

    if len(catalog_data) == 0:
        logger.info("No Galex data found")
        return all_filters

    match = catalog_data.group_by("distance_arcmin")[0]

    for i, b in enumerate(GALEX_BANDS):
        mag_col = GALEX_MAG_COLS[i]
        magerr_col = GALEX_MAGERR_COLS[i]

        mag = match[mag_col]
        magerr = match[magerr_col]

        if not np.ma.is_masked(mag):

            entry = Photometry.from_position(
                src_position=src_position,
                filter_name=f"galex_{b}",
                observed_mag=mag,
                mag_err=magerr,
            )
            all_filters.append(entry)

    logger.info(f"GALEX data found with {len(all_filters)} filters")

    return all_filters



"""
This module provides functionality to analyse the results of the fitting process
"""

from galsynthspec.datamodels.fitresult import FitResult
from galsynthspec.datamodels.galaxy import Galaxy
from galsynthspec.plotting.corner import plot_corner
from galsynthspec.plotting.sed import generate_sed_plot
from galsynthspec.utils.predict import get_predicted_photometry


def analyse_results(galaxy: Galaxy, res: FitResult):
    """
    Analyse the results of the fitting process and plot the corner plot.

    :param galaxy: Galaxy The galaxy object to analyse results for.
    :param res: Result The result of the fitting process.
    :return: None
    """

    plot_corner(res=res, out_path=galaxy.corner_path)
    sample_df = generate_sed_plot(res=res, out_dir=galaxy.base_output_dir)
    get_predicted_photometry(galaxy, res, sample_df=sample_df)



"""
Base Model for source
"""

import logging
from pathlib import Path

import numpy as np
import pandas as pd
from numpydantic import NDArray, Shape
from prospect.io import read_results as reader
from prospect.models import SpecModel
from prospect.plotting.utils import sample_posterior
from prospect.sources import CSPSpecBasis
from pydantic import BaseModel, ConfigDict, Field, model_validator
from sedpy.observate import Filter

from galsynthspec.model import get_model, get_sps

logger = logging.getLogger(__name__)


def weighted_quantiles(
    values: np.ndarray, weights: np.ndarray, quantiles=0.5
) -> np.ndarray:
    """
    Calculate weighted quantiles of a set of values.

    :param values: 2D array of values for which to calculate the quantiles
    :param weights:  Weights for the values, must be the same length as values
    :param quantiles: Quantiles to calculate, can be a single value or a list of values
    :return: Quantiles of the values, same length as values axis
    """
    i = np.argsort(values)
    c = np.cumsum(weights[i])
    return values[i[np.searchsorted(c, np.array(quantiles) * c[-1])]]


class BestFit(BaseModel):
    """
    Base model for best fit parameters
    """

    parameter: NDArray[Shape["5-6 x"], float] = Field(description="Best fit parameters")
    photometry: NDArray[Shape["* x"], float] = Field(description="Best fit photometry")
    restframe_wavelengths: NDArray[Shape["* x"], float] = Field(
        description="Restframe wavelengths for the best fit"
    )
    spectrum: NDArray[Shape["* x"], float] = Field(description="Best fit spectrum")
    mfrac: float = Field()

    @model_validator(mode="after")
    def validate_length(self):
        """
        Validate the parameter shape
        """
        if not len(self.restframe_wavelengths) == len(self.spectrum):
            raise ValueError(
                "Restframe wavelengths and spectrum must have the same length"
            )
        return self


class PredictedPhotometry(BaseModel):
    """
    Base model for predicted photometry
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    maggies: NDArray[Shape["* x"], float] = Field(
        description="Predicted photometry in maggies"
    )
    maggies_unc: NDArray[Shape["* x"], float] = Field(
        description="Uncertainty in the predicted photometry in maggies"
    )
    phot_mask: NDArray[Shape["* x"], bool] = Field(
        description="Mask for the photometry"
    )
    filternames: list[str] = Field(
        description="Names of the filters used for the photometry"
    )
    filters: list[Filter] = Field(
        description="List of sedpy filters used for the photometry"
    )


class FitResult(BaseModel):
    """
    Base model for fitting result
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    input_path: Path = Field(description="Path to input file")
    fit_parameters: list[str] = Field(description="Names of fit parameters")
    chain: NDArray[Shape["* x, 5-6 y"], float] = Field(
        description="The chain of the fit"
    )
    weights: NDArray[Shape["* x"], float] = Field(description="The weights of the fit")
    redshift: float = Field(description="The redshift of the source")
    model: SpecModel = Field(description="The model used for fitting")
    obs: dict = Field(description="The observation data")
    sps: CSPSpecBasis = Field(description="The SPS model used for fitting")
    best_fit: BestFit = Field(description="The best fit model")
    predicted_photometry: PredictedPhotometry = Field(
        description="Predicted photometry from the model"
    )

    @classmethod
    def from_file(cls, file_path: Path) -> "FitResult":
        """
        Read the result from a file

        :param file_path: Path to the file
        :return: Result instance
        """
        out, out_obs, _ = reader.results_from(str(file_path))
        model = get_model(redshift=out["obs"]["redshift"])
        return cls(
            input_path=file_path,
            fit_parameters=out["theta_labels"],
            chain=out["chain"][500:],
            weights=out["weights"][500:],
            redshift=out["obs"]["redshift"],
            obs=out["obs"],
            model=model,
            sps=get_sps(),
            best_fit=BestFit(**out["bestfit"]),
            predicted_photometry=PredictedPhotometry(**out_obs),
        )

    @model_validator(mode="after")
    def validate_chain(self):
        """
        Validate the source name
        """
        assert self.chain.shape[1] == len(  # pylint: disable=no-member
            self.fit_parameters
        )
        return self

    def sample_from_posterior(self, n_sample: int = 1) -> np.ndarray:
        """
        Sample from the posterior

        :param n_sample: Number of samples to draw
        :return: The samples
        """
        return sample_posterior(self.chain, weights=self.weights, nsample=n_sample)

    def sample_sed_from_posterior(self, n_sample: int = 1) -> pd.DataFrame:
        """
        Sample the SED from the posterior

        :param n_sample: Number of samples to draw
        :return: The sampled SEDs
        """
        all_pred = []

        thetas = self.sample_from_posterior(n_sample=n_sample)
        logger.info(f"Generating {n_sample} predictions from the posterior samples")
        for theta in thetas:
            spec, _, _ = self.predict(theta)
            all_pred.append(spec)

        all_pred = np.array(all_pred)
        return pd.DataFrame(all_pred)

    def predict(self, theta, obs=None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Predict the spectrum, photometry and mass fraction

        :param theta: Parameters to simulate
        :param obs: Set of photometric datapoints
        :return: Simulated spectrum, photometry and mass fraction
        """
        if obs is None:
            obs = self.obs

        spec, phot, mfrac = self.model.predict(  # pylint: disable=no-member
            theta, obs=obs, sps=self.sps
        )
        return spec, phot, mfrac

    @property
    def rest_frame_wavelengths(self) -> np.ndarray:
        """
        Get the restframe wavelengths for the best fit
        """
        return self.best_fit.restframe_wavelengths  # pylint: disable=no-member

    def get_redshift(self) -> float:
        """
        Get the redshift of the source

        :return: Redshift
        """
        redshift = self.redshift
        if redshift is None:
            idx = self.fit_parameters.index("zred")  # pylint: disable=no-member
            redshift = weighted_quantiles(
                self.chain[:, idx],
                self.weights,
                quantiles=0.5,
            )
            logger.info(f"Redshift is not known. Using median fit value: {redshift}")

        return redshift



"""
Model
"""

from prospect.sources import CSPSpecBasis


def get_sps() -> CSPSpecBasis:
    """
    Get the stellar population synthesis model

    :return: Stellar population synthesis model
    """

    return CSPSpecBasis(zcontinuous=1)



"""
Module for querying the SkyPortal API for source information.
"""

import logging

import pandas as pd

from galsynthspec.skyportal.base_client import client

logger = logging.getLogger(__name__)


def strip_tns_name(name: str) -> str:
    """
    Strip the TNS prefix from a name.

    :param name: The TNS name.
    :return: The stripped name.
    """
    is_digit = [x.isdigit() for x in name]
    idx = is_digit.index(True)
    tns_root = name[idx:].strip()
    return tns_root


def query_skyportal_by_name(name: str) -> pd.Series:
    """
    Query the SkyPortal API for a source by ZTF or AT name.

    :param name: AT/SN/TDE or ZTF name of the source.
    :return: Pandas Series with source data.
    """

    if not "ZTF" in name[:3]:
        # Must be TNS name, try to strip it
        name = strip_tns_name(name)

    res = client.api("GET", f"sources/{name}")
    res.raise_for_status()

    data = res.json()["data"]

    return pd.Series(data)



"""
Module for defining paths to data files and directories.
"""

import logging
import os
from pathlib import Path

from dotenv import load_dotenv

logger = logging.getLogger(__name__)

load_dotenv()

_data_dir = os.getenv("GALSYNTHSPEC_DATA_DIR", None)


if _data_dir is None:
    data_dir = Path.home() / "Data/galsynthspec"
    logger.warning(
        f"Environment variable GALSPECSYNTH_DATA_DIR not set. "
        f"Using default data directory {data_dir}."
    )

else:
    data_dir = Path(_data_dir)
    logger.info(
        f"Using data directory {data_dir} set by "
        f"environment variable GALSYNTHSPEC_DATA_DIR."
    )


sfd_path = Path(
    os.getenv(
        "SFDMAP_DATA_DIR",
        data_dir.joinpath("sfdmap/sfddata-master"),  # Default path for SFD map data
    )
)


def get_output_dir(source_name: str) -> Path:
    """
    Get the output directory for a given galaxy name.

    :param source_name: Name of the galaxy
    :return: Path to the output directory
    """
    out_dir = data_dir / source_name
    out_dir.mkdir(parents=True, exist_ok=True)
    return out_dir



"""
Module for WISE data
"""

import logging

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.ipac.irsa import Irsa

from galsynthspec.datamodels.photometry import Photometry

logger = logging.getLogger(__name__)

# WISE values from
# https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec4_4h.html
zeromag_wise = {
    "w1": 309.5 * u.Jansky,  # pylint: disable=no-member
    "w2": 171.8 * u.Jansky,  # pylint: disable=no-member
    "w3": 31.7 * u.Jansky,  # pylint: disable=no-member
    "w4": 8.4 * u.Jansky,  # pylint: disable=no-member
}
offsets_wise = {key: zm.to("mag(AB)").value for key, zm in zeromag_wise.items()}


def download_wise_data(
    src_position: SkyCoord,
    radius_arcsec: float,
) -> list[Photometry]:
    """
    Module for downloading WISE data for a given galaxy.

    Parameters
    :param src_position: SkyCoord The position of the source in the sky.
    :param radius_arcsec: float The radius of the search in arcseconds.

    Returns
    :return: list[Photometry] The photometry data for the source.
    """

    all_filters = []

    allwise = Irsa.query_region(
        src_position,
        catalog="allwise_p3as_psd",
        radius=radius_arcsec * u.arcsec,  # pylint: disable=no-member
    )

    if len(allwise) == 0:
        logger.info("No WISE data found")
        return all_filters

    match = allwise[0]

    for band in zeromag_wise:
        mag_raw = match[f"{band}mpro"]
        offset = offsets_wise[band]
        mag = mag_raw + offset
        mag_err = match[f"{band}sigmpro"]

        if np.ma.is_masked(match[f"{band}sigmpro"]):
            mag_err = mag
            mag = np.nan

        entry = Photometry.from_position(
            src_position=src_position,
            filter_name=f"wise_{band}",
            observed_mag=mag,
            mag_err=mag_err,
            vega_mag=mag_raw,
        )
        all_filters.append(entry)

    logger.info(f"WISE data found with {len(all_filters)} filters")

    return all_filters



"""
Client for interacting with Skyportal API
"""

import logging
import os
from typing import Mapping, Optional
from urllib.parse import urljoin

import requests
from dotenv import load_dotenv
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

DEFAULT_TIMEOUT = 5  # seconds

logger = logging.getLogger(__name__)

load_dotenv()
BASE_SKYPORTAL_URL = os.getenv("SKYPORTAL_URL", "https://fritz.science/api/")


def read_skyportal_token() -> str:
    """
    Get Skyportal token from environment variable.

    :return: Skyportal token
    """
    return os.getenv("SKYPORTAL_TOKEN")


class TimeoutHTTPAdapter(HTTPAdapter):
    """
    HTTP adapter that sets a default timeout for all requests.
    """

    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, *args, **kwargs):
        """
        Send a request with a default timeout.

        :param request: request to send
        :param args: args to pass to super().send
        :param kwargs: kwargs to pass to super().send
        :return: response from request
        """
        try:
            timeout = kwargs.get("timeout")
            if timeout is None:
                kwargs["timeout"] = self.timeout
            return super().send(request, *args, **kwargs)
        except AttributeError:
            kwargs["timeout"] = DEFAULT_TIMEOUT
            return super().send(request, *args, **kwargs)


class SkyportalClient:
    """
    Basic Skyportal client class for executing functions
    """

    def __init__(
        self,
        base_url: str = BASE_SKYPORTAL_URL,
    ):
        self.base_url = base_url
        self._session = None
        self.session_headers = None

    def set_up_session(self):
        """
        Set up a session for sending requests to Skyportal.

        :return: None
        """
        # session to talk to SkyPortal
        self._session = requests.Session()
        self.session_headers = {
            "Authorization": f"token {self._get_skyportal_token()}",
            "User-Agent": "galsynthspec",
        }

        retries = Retry(
            total=5,
            backoff_factor=2,
            status_forcelist=[405, 429, 500, 502, 503, 504],
            allowed_methods=["HEAD", "GET", "PUT", "POST", "PATCH"],
        )
        adapter = TimeoutHTTPAdapter(timeout=5, max_retries=retries)
        self._session.mount("https://", adapter)
        self._session.mount("http://", adapter)

    def get_session(self) -> requests.Session:
        """
        Wrapper for getting the session.
        If the session is not set up, it will be set up.

        :return: Session
        """
        if self._session is None:
            self.set_up_session()

        return self._session

    def ping(self) -> bool:
        """
        Ping the SkyPortal API to check if it is reachable.

        :return: True if the API is reachable, False otherwise
        """
        try:
            response = self.api("GET", "config")
            return response.status_code == 200
        except requests.RequestException as e:
            logger.error(f"Error pinging SkyPortal API: {e}")
            return False

    @staticmethod
    def _get_skyportal_token() -> str:
        """
        Get Skyportal token from environment variable.

        :return: Skyportal token
        """
        token_skyportal = read_skyportal_token()

        if token_skyportal is None:
            err = (
                "No skyportal token specified. Run 'export SKYPORTAL_TOKEN=<token>' to "
                "set. The skyportal token will need to be specified manually "
                "for skyportal API queries."
            )
            logger.error(err)
            raise ValueError(err)

        return token_skyportal

    @staticmethod
    def has_skyportal_token() -> bool:
        """
        Check if the Skyportal token is set in the environment.

        :return: True if the token is set, False otherwise
        """
        return read_skyportal_token() is not None

    def api(
        self, method: str, endpoint: str, data: Optional[Mapping] = None
    ) -> requests.Response:
        """Make an API call to a SkyPortal instance

        headers = {'Authorization': f'token {self.token}'}
        response = requests.request(method, endpoint, json_dict=data, headers=headers)

        :param method: HTTP method
        :param endpoint: API endpoint
        :param data: JSON data to send
        :return: response from API call
        """
        method = method.lower()

        session = self.get_session()

        methods = {
            "head": session.head,
            "get": session.get,
            "post": session.post,
            "put": session.put,
            "patch": session.patch,
            "delete": session.delete,
        }

        if endpoint is None:
            raise ValueError("Endpoint not specified")
        if method not in ["head", "get", "post", "put", "patch", "delete"]:
            raise ValueError(f"Unsupported method: {method}")

        url = urljoin(self.base_url, endpoint)

        if method == "get":
            response = methods[method](
                url,
                params=data,
                headers=self.session_headers,
            )
        else:
            response = methods[method](
                url,
                json=data,
                headers=self.session_headers,
            )

        return response


client = SkyportalClient()



"""
Utilities for accessing the Transient Name Server (TNS) API.
"""

import logging
from io import StringIO

import pandas as pd
import requests
from astropy import units as u
from astropy.coordinates import SkyCoord

from galsynthspec.paths import get_output_dir
from galsynthspec.skyportal.query import strip_tns_name

logger = logging.getLogger(__name__)

# Some TNS bug prevents API queries that do not come from internet browsers
TNS_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 "
    "(KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"
}
BASE_TNS_URL = "https://www.wis-tns.org/search?"

TNS_CACHE_NAME = "tns_info.json"


def query_tns_by_name(
    source_name: str, internal_name_bool: bool = False
) -> pd.DataFrame:
    """
    Query the Transient Name Server (TNS) by name.

    :param source_name: Name of the transient to search for.
    :param internal_name_bool: Boolean whether the name is an internal survey name.
    :return: A pandas DataFrame containing the search results.
    """

    query_arg = "internal_name=" if internal_name_bool else "name="

    search_url = (
        f"{BASE_TNS_URL}{query_arg}{strip_tns_name(source_name)}"
        f"&include_frb=0&format=csv&page=0"
    )
    response = requests.get(search_url, headers=TNS_HEADERS, timeout=10)
    csv_data = StringIO(response.text)
    return pd.read_csv(csv_data)


def download_tns(source_name: str) -> pd.Series:
    """
    Download the TNS data for a given name.

    :param source_name: The name of the transient as listed in TNS.
    :return: The first row of the TNS search results as a pandas Series.
    """
    # Query TNS by name e.g "AT2020mni"
    df = query_tns_by_name(source_name)

    # If no results found, try querying by internal name e.g "ZTF20abkavqj"
    if len(df) == 0:
        df = query_tns_by_name(source_name, internal_name_bool=True)

    # If still no results, raise an error
    if len(df) == 0:
        logger.error(f"No TNS data found for {source_name}.")
        raise ValueError(f"No TNS data found for {source_name}.")

    if len(df.columns) == 1:
        if df.columns == ["<html>"]:
            if "Forbidden" in df.iloc[0, 0]:
                logger.error(
                    "TNS API access forbidden. "
                    "Please check your network connection or TNS API status."
                )
                raise ConnectionError("TNS API access forbidden.")

    res = df.iloc[0].copy()

    # Add coordinates in degrees
    c = SkyCoord(
        res["RA"], res["DEC"], unit=(u.hourangle, u.deg)  # pylint: disable=no-member
    )
    res["ra"], res["dec"] = c.ra.deg, c.dec.deg
    res["redshift"] = res["Redshift"] if "Redshift" in res else None

    return res


def get_tns_by_name(tns_name: str, use_cache: bool = True) -> pd.Series:
    """
    Get information about a transient from TNS.

    :param tns_name: The name of the transient as listed in TNS.
    :param use_cache: If True, use cached TNS data if available.
    :return: A dictionary containing the TNS data for the transient.
    """
    output_dir = get_output_dir(tns_name)
    tns_file = output_dir / TNS_CACHE_NAME

    if use_cache & tns_file.exists():
        logger.info(f"Loading cached TNS data for {tns_name}")
        res = pd.read_json(tns_file, typ="series")
    else:
        logger.info(f"Downloading TNS data for {tns_name}")
        res = download_tns(tns_name)
        logger.info(f"Saving TNS data to {tns_file}")
        res.to_json(tns_file)

    return res



"""
Plot the results of the fitting
"""

import logging
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats

from galsynthspec.datamodels.fitresult import FitResult

logger = logging.getLogger(__name__)


def generate_sed_plot(
    res: FitResult,
    out_dir: Path,
) -> pd.DataFrame:
    """
    Function to generate a plot of the fitting results

    :param res: Result
    :param out_dir: Output path
    :return: SED DataFrame with the predicted SEDs
    """

    obs_wavelengths = res.rest_frame_wavelengths * (1 + res.get_redshift())

    df = res.sample_sed_from_posterior(n_sample=1000)

    plt.figure()
    ax = plt.subplot(111)

    # Plot Median
    plt.plot(obs_wavelengths, df.quantile(0.5), linestyle="-")

    sigmas = np.linspace(0.0, 3.0, 50)

    for sigma in sigmas:
        upper_percentile = stats.norm.cdf(sigma)
        lower_percentile = 1.0 - upper_percentile

        plt.fill_between(
            obs_wavelengths,
            df.quantile(upper_percentile),
            df.quantile(lower_percentile),
            alpha=1.0 / len(sigmas),
            color="C1",
        )

    pwave = np.array([f.wave_effective for f in res.predicted_photometry.filters])
    # plot the data

    maggies = res.predicted_photometry.maggies
    mask = maggies > 0.0

    ax.plot(pwave, maggies, linestyle="", marker="o", color="k")
    ax.errorbar(
        pwave,
        maggies,
        yerr=res.predicted_photometry.maggies_unc,
        linestyle="",
        color="k",
        zorder=10,
    )
    ax.set_ylabel(r"$f_\nu$ (maggies)")
    ax.set_xlabel(r"$\lambda$ (AA)")
    ax.set_xlim(1e3, 5e5)
    ax.set_ylim(maggies[mask].min() * 0.1, maggies[mask].max() * 5)
    ax.set_yscale("log")
    ax.set_xscale("log")

    out_path = out_dir / "sed_plot.pdf"
    logger.info(f"Saving SED plot to {out_path}")
    plt.savefig(out_path, bbox_inches="tight", dpi=300.0)
    plt.close()

    new = pd.DataFrame()
    new["wavelength"] = obs_wavelengths
    new["flux"] = df.quantile(0.5)
    new["sigma"] = 0.5 * (df.quantile(0.84) - df.quantile(0.16))
    sed_path = out_dir / "synthetic_sed.json"
    logger.info(f"Saving synthetic SED to {sed_path}")
    new.to_json(sed_path)

    return df



"""
Module to actually run galaxy synthesis and spectral synthesis.
"""

import logging

import numpy as np
from prospect.fitting import fit_model, lnprobfn
from prospect.io import write_results as writer
from prospect.utils.obsutils import fix_obs

from galsynthspec.datamodels.fitresult import FitResult
from galsynthspec.datamodels.galaxy import Galaxy
from galsynthspec.model import get_model, get_sps

logger = logging.getLogger(__name__)


def fit_galaxy(galaxy: Galaxy, use_cache: bool = True):
    """
    Fit a galaxy model to the photometry data of a given galaxy.

    :param galaxy: Galaxy The galaxy object containing the photometry data.
    :param use_cache: Bool If True, use cached results if available.
    :return: None
    """

    photometry_list = galaxy.get_photometry(use_cache=use_cache)

    filters = [p.filter for p in photometry_list]
    maggies = np.array([p.maggies for p in photometry_list])

    magerr = np.array([p.mag_err_combined for p in photometry_list])

    obs = {
        "wavelength": None,
        "spectrum": None,
        "unc": None,
        "redshift": galaxy.redshift,
        "maggies": maggies,
        "maggies_unc": magerr * maggies / 1.086,
        "filters": filters,
    }

    obs = fix_obs(obs)

    model = get_model(redshift=galaxy.redshift)

    noise_model = (None, None)

    sps = get_sps()

    fitting_kwargs = {
        "nlive_init": 1000,
        "nested_method": "rwalk",
        "nested_target_n_effective": 1000,
        "nested_dlogz_init": 0.05,
    }

    output = fit_model(
        obs,
        model,
        sps,
        optimize=False,
        dynesty=True,
        lnprobfn=lnprobfn,
        noise=noise_model,
        **fitting_kwargs,
    )
    _, duration = output["sampling"]

    galaxy.mcmc_cache_file.unlink(missing_ok=True)
    writer.write_hdf5(
        str(galaxy.mcmc_cache_file),
        {},
        model,
        obs,
        output["sampling"][0],
        None,
        sps=sps,
        tsample=output["sampling"][1],
        toptimize=0.0,
    )

    logger.info(
        f"Prospector run complete for {galaxy.source_name} in {duration:.1f} seconds"
    )


def get_galaxy_results(galaxy: Galaxy, use_cache: bool = True) -> FitResult:
    """
    Generate synthetic spectra for a given galaxy.

    :param galaxy: Galaxy The galaxy object to generate spectra for.
    :param use_cache: bool Whether to refit the model even if a cache file exists.
                        Defaults to False.
    :return: Result The result of the fitting process,
                including the model and observations.
    """
    hfile = galaxy.mcmc_cache_file

    if hfile.exists() & use_cache:
        logger.info(f"Cache file {hfile} already exists, skipping fitting.")
    else:
        fit_galaxy(galaxy, use_cache=use_cache)

    return galaxy.load_results()



"""
Base Model for source
"""

import logging
from pathlib import Path

import pandas as pd
from astropy import units as u
from astropy.coordinates import SkyCoord
from pydantic import BaseModel, Field, model_validator

from galsynthspec.datamodels.fitresult import FitResult
from galsynthspec.datamodels.photometry import Photometry
from galsynthspec.download import download_all_data
from galsynthspec.paths import get_output_dir

logger = logging.getLogger(__name__)


class Galaxy(BaseModel):
    """
    Base model for galaxy data
    """

    source_name: str | None = Field(description="Name of the source", default=None)
    ra_deg: float = Field(
        description="Right ascension of the source in degrees", ge=0.0, le=360.0
    )
    dec_deg: float = Field(
        description="Declination of the source in degrees", ge=-90.0, le=90.0
    )
    redshift: None | float = Field(
        description="Redshift of the source", ge=0.0, allow_inf_nan=True
    )

    @model_validator(mode="after")
    def validate_source_name(self):
        """
        Validate the source name
        """
        if self.source_name is None:
            src_position = SkyCoord(self.ra_deg, self.dec_deg, unit="deg")
            ra_str = src_position.ra.to_string(
                unit=u.hour, sep="", precision=2, pad=True  # pylint: disable=no-member
            )
            dec_str = src_position.dec.to_string(
                unit=u.degree,  # pylint: disable=no-member
                sep="",
                precision=2,
                alwayssign=True,
                pad=True,
            )
            j_name = f"J{ra_str}{dec_str}"
            logger.info(f"Source name not provided. Using J2000 name {j_name}")
            self.source_name = j_name

        return self

    @property
    def base_output_dir(self) -> Path:
        """
        Get the base output directory for the source
        """
        return get_output_dir(source_name=self.source_name)

    @property
    def sky_coord(self) -> SkyCoord:
        """
        Get the sky coordinate for the source
        """
        return SkyCoord(self.ra_deg, self.dec_deg, unit="deg")

    @property
    def photometry_cache_file(self) -> Path:
        """
        Get the cache file for the photometry
        """
        return self.base_output_dir / "photometry.json"

    @property
    def mcmc_cache_file(self) -> Path:
        """
        Get the cache file for MCMC

        :return: Cache path
        """
        return self.base_output_dir / "quickstart_dynesty_mcmc_mod.h5"

    @property
    def synthetic_photometry_file(self) -> Path:
        """
        Get the synthetic photometry file

        :return: Synthetic photometry path
        """
        return self.base_output_dir / "synthetic_photometry.json"

    @property
    def corner_path(self) -> Path:
        """
        Get the cache file for MCMC

        :return: Cache path
        """
        return self.base_output_dir / "corner.pdf"

    def get_photometry(
        self, radius_arcsec: float = 3.0, use_cache: bool = True
    ) -> list[Photometry]:
        """
        Get the photometry data for the source

        :param radius_arcsec: float The radius of the search in arcseconds
        :param use_cache: bool If True, use the cached photometry data if available

        :return: list[Photometry] The photometry data
        """
        if self.photometry_cache_file.is_file() and use_cache:
            return self.load_photometry_from_cache()

        photometry = download_all_data(self.sky_coord, radius_arcsec=radius_arcsec)
        self.export_photometry_to_cache(photometry)

        return photometry

    def export_photometry_to_cache(self, photometry: list[Photometry]):
        """
        Export the photometry to a cache file

        :param photometry: list[Photometry] The photometry data

        :return: None
        """
        logger.info(f"Exporting photometry to {self.photometry_cache_file}")
        pd.DataFrame([p.model_dump() for p in photometry]).to_json(
            self.photometry_cache_file
        )

    def load_photometry_from_cache(self) -> list[Photometry]:
        """
        Load the photometry from the cache file

        :return: List of Photometry objects loaded from the cache file
        """
        logger.info(f"Loading photometry from cache file {self.photometry_cache_file}")
        df = pd.read_json(self.photometry_cache_file)
        return [Photometry.model_validate(p) for p in df.to_dict(orient="records")]

    def load_results(self) -> FitResult:
        """
        Load the results for the source

        :return: Result object containing the results
        """
        if not self.mcmc_cache_file.is_file():
            raise FileNotFoundError(
                f"MCMC cache file {self.mcmc_cache_file} does not exist."
            )

        logger.info(f"Loading results from {self.mcmc_cache_file}")
        return FitResult.from_file(self.mcmc_cache_file)



"""
Module to iteratively download photometry data for a given galaxy.
"""

from astropy.coordinates import SkyCoord

from galsynthspec.datamodels.photometry import Photometry
from galsynthspec.download.galex import download_galex_data
from galsynthspec.download.ps1 import download_ps1_data
from galsynthspec.download.sdss import download_sdss_data
from galsynthspec.download.twomass import download_twomass_data
from galsynthspec.download.wise import download_wise_data


def download_all_data(
    src_position: SkyCoord,
    radius_arcsec: float,
) -> list[Photometry]:
    """
    Module for downloading photometry data for a given galaxy.

    :param src_position: SkyCoord The position of the source in the sky.
    :param radius_arcsec: Radius of the search in arcseconds.
    :return: Returns a list of Photometry objects.
    """

    # Optical data
    all_filters = download_sdss_data(src_position, radius_arcsec)
    # Download PS1 if SDSS is not available
    if len(all_filters) == 0:
        all_filters = download_ps1_data(src_position, radius_arcsec)

    # UV data
    all_filters.extend(download_galex_data(src_position, radius_arcsec))
    # NIR data
    all_filters.extend(download_twomass_data(src_position, radius_arcsec))
    # MIR data
    all_filters.extend(download_wise_data(src_position, radius_arcsec))

    return all_filters



"""
This module predicts the photometry for a galaxy
based on the results of a fitting procedure.
"""

import pandas as pd
from prospect.sources.constants import jansky_cgs, lightspeed
from scipy import stats
from sedpy.observate import getSED, load_filters

from galsynthspec.datamodels.fitresult import FitResult
from galsynthspec.datamodels.galaxy import Galaxy
from galsynthspec.utils.extinction import get_extinction_for_filter

DEFAULT_FILTER_LIST = [
    "galex_FUV",
    "galex_NUV",
    "uvot_w2",
    "uvot_m2",
    "uvot_w1",
    "sdss_u0",
    "sdss_g0",
    "sdss_r0",
    "sdss_i0",
    "sdss_z0",
    "twomass_J",
    "twomass_H",
    "twomass_Ks",
    "wise_w1",
    "wise_w2",
    "wise_w3",
    "wise_w4",
]


def get_lambda_cgs(x, angstroms):
    """
    Convert a wavelength in Angstroms to a wavelength in CGS units (cm).

    :param x: Flux in arbitrary units (e.g., erg/s/cm^2/Angstrom).
    :param angstroms: Array of wavelengths in Angstroms.
    :return: Flux in CGS units (erg/s/cm^2/nm).
    """
    return x * lightspeed / angstroms**2.0 * (3631 * jansky_cgs)


def get_lambda_quantile(df, q, angstroms):
    """
    Get the quantile of the wavelength in CGS units from a DataFrame.

    :param df: DataFrame containing the sampled SEDs.
    :param q: Quantile to compute (e.g., 0.5 for median).
    :param angstroms: Array of wavelengths in Angstroms.
    :return: Quantile flux in CGS units (erg/s/cm^2/nm).
    """
    return get_lambda_cgs(df.quantile(q).to_numpy(), angstroms)


def get_photometry_quantile(angstroms, df, q, filters: list[str]):
    """
    Get the photometry quantile for a given set of wavelengths and a DataFrame of SEDs.

    :param angstroms: Array of wavelengths in Angstroms.
    :param df: DataFrame containing the sampled SEDs.
    :param q: Quantile to compute (e.g., 0.5 for median).
    :param filters: List of filter names to compute the photometry for.
    :return: Magnitudes corresponding to the specified quantile for each filter.
    """
    filterlist = load_filters(filters)
    f_lambda_cgs = get_lambda_quantile(df, q, angstroms)
    mags = getSED(angstroms, f_lambda_cgs, filterlist=filterlist)
    return mags


def get_predicted_photometry(
    galaxy: Galaxy,
    result: FitResult,
    sample_df: pd.DataFrame | None = None,
    filter_list: None | list[str] = None,
) -> pd.DataFrame:
    """
    Function to get the predicted photometry for a galaxy based
    on the result of a fitting procedure.

    :param galaxy: Galaxy
    :param result: Result of the MCMC fitting procedure.
    :param sample_df: DataFrame containing the sampled SEDs from the posterior.
                        If None, it will sample 1000 SEDs.
    :param filter_list: List of filters to predict photometry for.
                        If None, it will use a default list of filters.
    :return: pd.DataFrame containing the predicted photometry.
    """

    if sample_df is None:
        sample_df = result.sample_sed_from_posterior(n_sample=1000)

    angstroms = result.rest_frame_wavelengths * (1.0 + result.get_redshift())

    if filter_list is None:
        filter_list = DEFAULT_FILTER_LIST

    photometry_dict = {x.filter_name: x for x in galaxy.get_photometry()}

    sigma = 1.0

    upper_percentile = stats.norm.cdf(sigma)
    lower_percentile = 1.0 - upper_percentile

    med = get_photometry_quantile(angstroms, sample_df, 0.5, filters=filter_list)
    up_pred = get_photometry_quantile(
        angstroms, sample_df, upper_percentile, filters=filter_list
    )
    lower_pred = get_photometry_quantile(
        angstroms, sample_df, lower_percentile, filters=filter_list
    )

    phot_df = pd.DataFrame(
        {
            "band": filter_list,
            "predicted_mag": med,
            "sigma+": med - up_pred,
            "sigma-": lower_pred - med,
            "measured_mag": [
                photometry_dict[x].observed_mag if x in photometry_dict else None
                for x in filter_list
            ],
            "measured_err": [
                photometry_dict[x].mag_err if x in photometry_dict else None
                for x in filter_list
            ],
            "extinction": [
                get_extinction_for_filter(galaxy.sky_coord, x) for x in filter_list
            ],
            "measured_mag_deextincted": [
                photometry_dict[x].mag if x in photometry_dict else None
                for x in filter_list
            ],
        }
    )
    phot_df["predicted_mag_extincted"] = (
        phot_df["predicted_mag"] + phot_df["extinction"]
    )

    print(phot_df)

    phot_df.to_json(galaxy.synthetic_photometry_file)
    return phot_df



"""
Module for PS1 data
"""

import logging

from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.mast import Catalogs

from galsynthspec.datamodels.photometry import Photometry

PS1_BANDS = ["g", "r", "i", "z"]
PS1_MAG_COLS = [f"{b}MeanKronMag" for b in PS1_BANDS]
PS1_MAGERR_COLS = [f"{b}MeanKronMagStd" for b in PS1_BANDS]

logger = logging.getLogger(__name__)


def download_ps1_data(
    src_position: SkyCoord,
    radius_arcsec: float,
) -> list[Photometry]:
    """
    Module for downloading PS1 data for a given galaxy.

    :param src_position: SkyCoord The position of the source in the sky.
    :param radius_arcsec: float The radius of the search in arcseconds.

    :return: list[Photometry] The photometry data for the source.
    """

    all_filters = []

    catalog_data = Catalogs.query_region(  # pylint: disable=no-member
        src_position,
        radius=radius_arcsec * u.arcsec,  # pylint: disable=no-member
        catalog="Panstarrs",
    )

    if len(catalog_data) == 0:
        logger.info("No Panstarrs data found")
        return all_filters

    match = catalog_data.group_by("distance")[0]

    for i, b in enumerate(PS1_BANDS):
        mag_col = PS1_MAG_COLS[i]
        magerr_col = PS1_MAGERR_COLS[i]

        mag = match[mag_col]
        mag_err = match[magerr_col]

        filter_name = f"ps1::{b}" if b in ["y"] else f"sdss_{b}0"

        entry = Photometry.from_position(
            src_position=src_position,
            filter_name=filter_name,
            observed_mag=mag,
            mag_err=mag_err,
        )
        all_filters.append(entry)

    logger.info(f"PS1 data found with {len(all_filters)} filters")

    return all_filters



"""
Module for running the galaxy synthetic spectra pipeline.
"""

from galsynthspec.datamodels.galaxy import Galaxy
from galsynthspec.run.analyse import analyse_results
from galsynthspec.run.fit import get_galaxy_results


def run_on_galaxy(galaxy: Galaxy, use_cache: bool = True):
    """
    Run the galaxy synthetic spectra pipeline for a given galaxy.

    :param galaxy: Galaxy The galaxy object to run the pipeline on.
    :param use_cache: bool Whether to use cached results if available.
    :return:
    """

    res = get_galaxy_results(galaxy, use_cache=use_cache)
    analyse_results(galaxy, res)



"""
Create a corner plot of the chain
"""

import logging
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from prospect.plotting import corner

from galsynthspec.datamodels.fitresult import FitResult, weighted_quantiles

logger = logging.getLogger(__name__)


def plot_corner(res: FitResult, out_path: Path):
    """
    Plot the corner plot of the chain

    :param res: The result of the fitting
    :param out_path: The output of the sampling
    """
    _, ndim = res.chain.shape
    cfig, axes = plt.subplots(ndim, ndim, figsize=(10, 9))
    corner.allcorner(
        res.chain.T,
        res.fit_parameters,
        axes,
        weights=res.weights,
        color="royalblue",
        show_titles=True,
    )
    plt.savefig(out_path, bbox_inches="tight")
    plt.close(cfig)

    logger.info(f"Corner plot saved to {out_path}")

    results = []
    for i, param in enumerate(res.fit_parameters):

        logify = param in ["mass"]

        vals = res.chain[:, i] if not logify else np.log10(res.chain[:, i])

        quantiles = weighted_quantiles(vals, res.weights, [0.16, 0.5, 0.84])
        results.append(
            {
                "parameter": param if not logify else f"log10({param})",
                "median": quantiles[1],
                "sigma-": quantiles[1] - quantiles[0],
                "sigma+": quantiles[2] - quantiles[1],
            }
        )

    df = pd.DataFrame(results)
    print(df)
    df.to_json(out_path.parent / "fit_results.json")



"""
CLI wrapper for galaxy synthetic spectra
"""

import logging

import click

from galsynthspec.datamodels.galaxy import Galaxy
from galsynthspec.run import run_on_galaxy
from galsynthspec.utils.query import query_by_name

logger = logging.getLogger(__name__)

logging.basicConfig(level=logging.INFO)


@click.group()
def cli():
    """
    CLI for galaxy synthetic spectra.
    """


@cli.command("by-name")
@click.argument("name", type=str)
@click.option(
    "--use-cache/--no-cache", default=True, help="Enable using cached results"
)
@click.option("-z", "--redshift", type=float, default=None)
def run_by_name(name, use_cache: bool, redshift: float = None):
    """
    Run the galaxy synthetic spectra pipeline for a given galaxy name.
    """
    logger.info(f"Running pipeline for source name {name}")
    gal = query_by_name(name)
    if gal.redshift is None:
        gal.redshift = redshift
    run_on_galaxy(gal, use_cache=use_cache)


@cli.command("by-ra-dec")
@click.argument("ra_deg", type=float)
@click.argument("dec_deg", type=float)
@click.option("-n", "--name", type=str, default=None)
@click.option("-z", "--redshift", type=float, default=None)
def run_by_ra_dec(ra_deg: float, dec_deg: float, name=None, redshift=None):
    """
    Run the galaxy synthetic spectra pipeline for a given galaxy name.
    """
    logger.info(f"Running pipeline for position {ra_deg} {dec_deg}")

    gal = Galaxy(source_name=name, ra_deg=ra_deg, dec_deg=dec_deg, redshift=redshift)

    run_on_galaxy(gal)



"""
Module for SDSS data
"""

import logging

from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.sdss import SDSS

from galsynthspec.datamodels.photometry import Photometry

SDSS_BANDS = ["u", "g", "r", "i", "z"]
SDSS_MAG_COLS = [f"cModelMag_{b}" for b in SDSS_BANDS]
SDSS_MAGERR_COLS = [f"cModelMagErr_{b}" for b in SDSS_BANDS]

logger = logging.getLogger(__name__)


def download_sdss_data(
    src_position: SkyCoord,
    radius_arcsec: float,
) -> list[Photometry]:
    """
    Module for downloading SDSS data for a given galaxy.

    Parameters
    :param src_position: SkyCoord The position of the source in the sky.
    :param radius_arcsec: float The radius of the search in arcseconds.

    Returns
    :return: list[Photometry] The photometry data for the source.
    """

    all_filters = []

    cat = SDSS.query_crossid(  # pylint: disable=no-member
        src_position,
        radius=radius_arcsec * u.arcsec,  # pylint: disable=no-member
        photoobj_fields=["ra", "dec"] + SDSS_MAG_COLS + SDSS_MAGERR_COLS,
    )

    if cat is None:
        logger.info("No SDSS data found")
        return all_filters

    for i, b in enumerate(SDSS_BANDS):
        mag_col = SDSS_MAG_COLS[i]
        magerr_col = SDSS_MAGERR_COLS[i]

        mag = cat[0][mag_col]
        magerr = cat[0][magerr_col]

        entry = Photometry.from_position(
            src_position=src_position,
            filter_name=f"sdss_{b}0",
            observed_mag=mag,
            mag_err=magerr,
        )
        all_filters.append(entry)

    logger.info(f"SDSS data found with {len(all_filters)} filters")

    return all_filters



"""
Module for calculating extinction from the
Schlegel, Finkbeiner, and Davis (1998) dust maps.
"""

import logging

import extinction
import numpy as np
from astropy.coordinates import SkyCoord
from sedpy.observate import load_filters
from sfdmap2 import sfdmap

from galsynthspec.paths import sfd_path

logger = logging.getLogger(__name__)

m = sfdmap.SFDMap(sfd_path.as_posix())


def get_extinction_correction(
    ra_deg: float, dec_deg: float, wavelengths_angstroms: list[float]
) -> list[float]:
    """
    Apply extinction correction

    See ... citation
    """
    coordinates = SkyCoord(ra_deg, dec_deg, frame="icrs", unit="degree")
    ebv = m.ebv(coordinates)
    wave = np.array(wavelengths_angstroms)

    return extinction.fitzpatrick99(  # pylint: disable=c-extension-no-member
        wave, 3.1 * ebv
    )


def get_extinction_for_filter(
    src_position: SkyCoord,
    filter_name: str,
) -> float:
    """
    Get the extinction correction for a given filter at the source position.

    :param src_position: SkyCoord The position of the source in the sky.
    :param filter_name: Name of the filter to get the extinction for.
    :return: Float The extinction correction value for the filter.
    """
    res = load_filters([filter_name])[0]

    mean = np.average(res.wavelength, weights=res.transmission)

    extinction_value = get_extinction_correction(
        ra_deg=src_position.ra.deg,
        dec_deg=src_position.dec.deg,
        wavelengths_angstroms=[mean],
    )[0]
    return extinction_value



"""
Module for generating base prospector model
"""

from prospect.models import SpecModel
from prospect.models.priors import LogUniform, Uniform
from prospect.models.templates import TemplateLibrary


def get_model(redshift: float | None = None) -> SpecModel:
    """
    Get the base prospector model

    :param redshift: Will fix the redshift if provided
    :return: SpecModel
    """

    model_params = TemplateLibrary["parametric_sfh"]
    model_params.update(TemplateLibrary["nebular"])

    if redshift is not None:
        # Fix redshift
        model_params["zred"]["init"] = redshift
    else:
        # Change redshift to free parameter
        model_params["zred"]["isfree"] = True
        model_params["zred"]["prior"] = LogUniform(mini=0.0, maxi=4.0)

    model_params["tage"]["prior"] = Uniform(mini=0.1, maxi=10.1)
    model_params["tau"]["prior"] = Uniform(mini=0.1, maxi=10.0)
    model_params["logzsol"]["prior"] = Uniform(mini=-1.8, maxi=0.2)
    model_params["dust2"]["prior"] = Uniform(mini=0.0, maxi=1.0)

    model = SpecModel(model_params)
    return model



"""
Base Model for photometry data
"""

import numpy as np
from pydantic import BaseModel, Field, model_validator
from sedpy.observate import Filter, list_available_filters

from galsynthspec.utils.extinction import get_extinction_for_filter


class Photometry(BaseModel):
    """
    Base model for photometry data
    """

    filter_name: str = Field(description="Name of the filter")
    observed_mag: float = Field(description="AB Magnitude for the photometry")
    extinction: float = Field(description="Extinction factor for the photometry")
    vega_mag: float | None = Field(
        description="Vega Magnitude for the photometry", default=None
    )
    mag_err: float = Field(description="Error in the photometry")
    systematic_error: float = Field(
        description="Systematic error in the photometry", default=0.05
    )

    @property
    def mag(self) -> float:
        """
        Get the magnitude in AB system
        """
        return self.observed_mag - self.extinction

    @model_validator(mode="after")
    def validate_filter(self):
        """
        Validate the filter
        """
        if self.filter_name not in list_available_filters():
            raise ValueError(
                f"Filter {self.filter_name} not found. "
                f"Available filters are {list_available_filters()}"
            )

        return self

    @property
    def filter(self) -> Filter:
        """
        Get the filter object for the photometry
        """
        return Filter(self.filter_name)

    @property
    def maggies(self) -> float:
        """
        Convert the magnitude to maggies
        """
        return 10.0 ** (-0.4 * self.mag) if not np.isnan(self.observed_mag) else 0.0

    @property
    def mag_err_combined(self) -> float:
        """
        Get the systematic error in the magnitude
        """
        return np.hypot(self.mag_err, self.systematic_error)

    @classmethod
    def from_position(cls, src_position, filter_name: str, **kwargs) -> "Photometry":
        """
        Create a Photometry instance from source position and filter data
        """
        extinction = get_extinction_for_filter(
            src_position=src_position, filter_name=filter_name
        )
        return cls(filter_name=filter_name, extinction=extinction, **kwargs)



"""
Module for testing galsynthspec
"""

import logging
import unittest
from pathlib import Path

import pandas as pd
from click.testing import CliRunner

from galsynthspec.cli.wrappers import cli
from galsynthspec.datamodels.galaxy import Galaxy

logger = logging.getLogger(__name__)

expected_df = pd.read_json(
    Path(__file__).parent / "test_data/synthetic_photometry.json"
)


class TestGalsynthspec(unittest.TestCase):
    """
    Class for testing Galsynthspec
    """

    def test_by_name(self):
        """
        Test ping

        :return: None
        """

        test_ra, test_dec = 314.262256, 14.204368
        redshift = 0.0512

        logger.info(f"Testing Galsynthspec for {test_ra}, {test_dec}")

        runner = CliRunner()
        runner.invoke(
            cli,
            ["by-ra-dec", str(test_ra), str(test_dec), "-z", str(redshift)],
            catch_exceptions=False,
        )

        # Check results

        logger.info("Checking results")

        gal = Galaxy(
            ra_deg=test_ra,
            dec_deg=test_dec,
            redshift=redshift,
        )
        assert gal.synthetic_photometry_file.exists()
        new = pd.read_json(gal.synthetic_photometry_file)

        pred_diff = (new["predicted_mag"] - expected_df["predicted_mag"]) / (
            0.5 * (expected_df["sigma+"] + expected_df["sigma-"])
        )

        max_diff = pred_diff.abs().max()
        logger.info(f"Max diff: {max_diff}")
        assert max_diff < 0.7, f"Difference too large: {max_diff}"

        median_diff = abs(pred_diff.median())
        logger.info(f"Median diff: {median_diff}")
        assert median_diff < 0.15, f"Median difference too large: {median_diff}"

        pd.testing.assert_series_equal(new["measured_mag"], expected_df["measured_mag"])
        pd.testing.assert_series_equal(
            new["extinction"], expected_df["extinction"], check_exact=False, atol=0.001
        )



"""
Module for 2MASS data
"""

import logging

from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.gaia import Gaia
from astroquery.irsa import Irsa

from galsynthspec.datamodels.photometry import Photometry

# Silence astroquery verbiage
logging.getLogger("astroquery").setLevel(logging.WARNING)

logger = logging.getLogger(__name__)

# 2MASS values from https://iopscience.iop.org/article/10.1086/376474
zeromag_2mass = {
    "J": 1594.0 * u.Jansky,  # pylint: disable=no-member
    "H": 1024.0 * u.Jansky,  # pylint: disable=no-member
    "Ks": 666.8 * u.Jansky,  # pylint: disable=no-member
}
offsets_2mass = {key: zm.to("mag(AB)").value for key, zm in zeromag_2mass.items()}


def download_twomass_extended(
    src_position: SkyCoord,
    radius_arcsec: float,
) -> list[Photometry]:
    """
    Module for downloading 2MASS extended data for a given galaxy.

    Parameters
    :param src_position: SkyCoord The position of the source in the sky.
    :param radius_arcsec: float The radius of the search in arcseconds.

    Returns
    :return: list[Photometry] The photometry data for the source.
    """

    all_filters = []

    extended_matches = Irsa.query_region(
        src_position,
        catalog="ext_src_cat",
        radius=radius_arcsec * u.arcsec,  # pylint: disable=no-member
    )

    if len(extended_matches) == 0:
        logger.info("No 2MASS extended data found")
        return all_filters

    match = extended_matches[0]

    for band in zeromag_2mass:
        mag_raw = match[f"{band.lower()[0]}_m_k20fe"]
        offset = offsets_2mass[band]
        # Convert from Vega to AB mag
        mag = mag_raw + offset
        magerr = match[f"{band.lower()[0]}_msig_k20fe"]

        entry = Photometry.from_position(
            src_position=src_position,
            filter_name=f"twomass_{band}",
            observed_mag=mag,
            mag_err=magerr,
            vega_mag=mag_raw,
        )
        all_filters.append(entry)

    logger.info(f"2MASS extended data found with {len(all_filters)} filters")

    return all_filters


def download_twomass_ps(
    src_position: SkyCoord,
    radius_arcsec: float,
) -> list[Photometry]:
    """
    Module for downloading 2MASS data for a given galaxy.

    Parameters
    :param src_position: SkyCoord The position of the source in the sky.
    :param radius_arcsec: float The radius of the search in arcseconds.

    Returns
    :return: list[Photometry] The photometry data for the source.
    """

    all_filters = []

    cmd = (
        f"SELECT * FROM gaiadr3.gaia_source AS g "
        f"JOIN gaiadr3.tmass_psc_xsc_best_neighbour AS xmatch USING (source_id) "
        f"JOIN gaiadr3.tmass_psc_xsc_join AS xjoin "
        f"  ON xmatch.original_ext_source_id = xjoin.original_psc_source_id "
        f"JOIN gaiadr1.tmass_original_valid AS tmass "
        f"  ON xjoin.original_psc_source_id = tmass.designation "
        f"WHERE CONTAINS(POINT('ICRS', g.ra, g.dec), "
        f"CIRCLE('ICRS', {src_position.ra.deg:.4f}, {src_position.dec.deg:.4f}, "
        f"{radius_arcsec / 3600.:.4f}))=1 "
        f"AND xmatch.number_of_mates=0 "
        f"AND xmatch.number_of_neighbours=1"
        f";"
    )

    job = Gaia.launch_job_async(cmd, dump_to_file=False)

    src_list = job.get_results()

    if len(src_list) == 0:
        logger.info("No 2MASS data found")
        return all_filters

    for band in zeromag_2mass:
        mag_raw = src_list[f"{band.lower()}_m"][0]
        offset = offsets_2mass[band]
        # Convert from Vega to AB mag
        mag = mag_raw + offset

        mag_err = src_list[f"{band.lower()}_msigcom"][0]

        entry = Photometry.from_position(
            src_position=src_position,
            filter_name=f"twomass_{band}",
            observed_mag=mag,
            mag_err=mag_err,
            vega_mag=mag_raw,
        )
        all_filters.append(entry)

    logger.info(f"2MASS data found with {len(all_filters)} filters")

    return all_filters


def download_twomass_data(
    src_position: SkyCoord,
    radius_arcsec: float,
) -> list[Photometry]:
    """
    Module for downloading 2MASS data for a given galaxy.

    Parameters
    :param src_position: SkyCoord The position of the source in the sky.
    :param radius_arcsec: float The radius of the search in arcseconds.

    Returns
    :return: list[Photometry] The photometry data for the source.
    """

    all_filters = download_twomass_extended(src_position, radius_arcsec)

    if len(all_filters) == 0:
        all_filters = download_twomass_ps(src_position, radius_arcsec)

    return all_filters
